from __future__ import annotations

import unittest
from dataclasses import dataclass
from datetime import datetime

from Aspidites._vendor.apm import *


class BasicUseCases(unittest.TestCase):

    def test_dict(self):
        self.assertTrue(match(
            {"foo": 1, "bar": 2, "baz": 3, "quux": 4},
            {"foo": 1}
        ))
        self.assertFalse(match(
            {"foo": 1, "bar": 2, "baz": 3, "quux": 4},
            Strict({"foo": 1})
        ))
        positive_number = InstanceOf(int) & Check(lambda v: v >= 0)
        self.assertTrue(match(
            {"foo": 1, "bar": 2},
            Strict({"foo": positive_number, "bar": positive_number})
        ))

    def test_tuple(self):
        self.assertTrue(match(
            (1, 2),
            (..., ...)
        ))

    def test_strict(self):
        self.assertTrue(match(1, 1.0))
        self.assertTrue(match(1.0, Strict(1.0)))
        self.assertTrue(match(1, Strict(1)))
        self.assertFalse(match(1, Strict(1.0)))

    def test_truish(self):
        self.assertTrue(match(1, IsTruish))
        self.assertTrue(match(2.9, IsTruish))
        self.assertFalse(match("", IsTruish))

    def test_length(self):
        self.assertTrue(match([], Length(exactly=0)))
        self.assertTrue(match(('abc',), Length(exactly=1)))
        self.assertTrue(match('abc', Length(exactly=3)))
        self.assertTrue(match('abc', Length(at_least=1, at_most=4)))
        self.assertFalse(match('abc', Length(at_least=4, at_most=10)))
        self.assertTrue(match('abcde', Length(at_least=4, at_most=5)))
        self.assertFalse(match('abcdef', Length(at_least=4, at_most=5)))

    def test_contains(self):
        self.assertTrue(match([1, 2, 3], Contains(2) & Contains(3)))
        self.assertTrue(match((1, 2, 3), Contains(2) & Contains(3)))
        self.assertTrue(match({1, 2, 3}, Contains(2) & Contains(3)))
        self.assertTrue(match({1: 'a', 2: 'b', 3: 'c'}, Contains(2) & Contains(3)))
        self.assertTrue(match("quux", Contains('uu')))

    def test_regex(self):
        self.assertTrue(match(
            "Hello, World!",
            Regex("[A-Z][a-z]+, [A-Z][a-z]+!")
        ))
        self.assertFalse(match(
            "Hello, World! How are you today?",
            Regex("[A-Z][a-z]+, [A-Z][a-z]+!")
        ))
        self.assertTrue(match(
            "Hello, World! How are you today?",
            Regex("[A-Z][a-z]+, [A-Z][a-z]+!.*")
        ))

    def test_remaining(self):
        self.assertTrue(match(
            [1, 2, 3, 4],
            [1, 2, 3, Remaining(InstanceOf(int), at_least=1)]))
        self.assertTrue(match(
            [1, 2, 3],
            [1, 2, 3, Remaining(InstanceOf(int))]))
        self.assertTrue(match(
            [1, 2, 3, 4, 5],
            [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 10))]))
        self.assertFalse(match(
            [1, 2, 3, 4, 5],
            [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 3))]))
        self.assertFalse(match(
            [1, 2, 3, 4],
            [1, 2, 3, Remaining(InstanceOf(int), at_least=2)]))

    def test_remaining_with_range(self):
        self.assertTrue(match(
            range(1, 5),
            [1, 2, 3, Remaining(InstanceOf(int), at_least=1)]))
        self.assertTrue(match(
            range(1, 4),
            [1, 2, 3, Remaining(InstanceOf(int))]))
        self.assertTrue(match(
            range(1, 5),
            [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 10))]))
        self.assertFalse(match(
            range(1, 5),
            [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 3))]))
        self.assertFalse(match(
            range(1, 5),
            [1, 2, 3, Remaining(InstanceOf(int), at_least=2)]))

    def test_capture(self):
        result = match(
            {
                "This": "Is A Rather Complex Beast",
                "Created-At": "Sun Jan  3 04:08:57 CET 2021",
                "User": {
                    "UserId": 102384,
                    "FirstName": "Jane",
                    "LastName": "Doe",
                }
            },
            {
                "User": Capture({
                    "FirstName": Capture(..., name="first_name"),
                    "LastName": Capture(..., name="last_name"),
                }, name="user")
            }
        )
        self.assertTrue(result)
        self.assertEqual("Jane", result['first_name'])
        self.assertEqual("Doe", result['last_name'])
        self.assertEqual(102384, result['user']['UserId'])

    def test_remaining_with_capture(self):
        result = match(
            [1, 2, 3, 4],
            [1, 2, Capture(Remaining(...), name="tail")]
        )
        self.assertTrue(result)
        self.assertEqual([3, 4], result['tail'])

    def test_and(self):
        self.assertFalse(match(
            0,
            Between(0, 1) & Between(1, 2)
        ))
        self.assertTrue(match(
            1,
            Between(0, 1) & Between(1, 2)
        ))
        self.assertFalse(match(
            2,
            Between(0, 1) & Between(1, 2)
        ))

    def test_or(self):
        self.assertTrue(match(
            0,
            Between(0, 1) | Between(1, 2)
        ))
        self.assertTrue(match(
            1,
            Between(0, 1) | Between(1, 2)
        ))
        self.assertTrue(match(
            2,
            Between(0, 1) | Between(1, 2)
        ))

    def test_xor(self):
        self.assertTrue(match(
            0,
            Between(0, 1) ^ Between(1, 2)
        ))
        self.assertFalse(match(
            1,
            Between(0, 1) ^ Between(1, 2)
        ))
        self.assertTrue(match(
            2,
            Between(0, 1) ^ Between(1, 2)
        ))

    def test_each(self):
        self.assertTrue(match([1, 2, 3], Each(InstanceOf(int))))
        self.assertFalse(match([1, 2.0, 3], Each(InstanceOf(int))))

    def test_each_item(self):
        pattern = EachItem(Regex("[a-z]+"), InstanceOf(int))
        self.assertTrue(match({"a": 1, "b": 2}, pattern))
        self.assertFalse(match({"_a": 1, "b": 2}, pattern))
        self.assertFalse(match({"a": 7.5}, pattern))

    def test_transformed(self):
        self.assertTrue(match("value", Transformed(len, 5)))

    def test_arguments(self):
        # noinspection PyUnusedLocal
        def f(x: int, y: str):
            pass

        # noinspection PyUnusedLocal
        def g(x):
            pass

        # noinspection PyUnusedLocal
        def h(x, y: int, m: str, n: str):
            pass

        self.assertTrue(match(f, Arguments(int, str)))
        self.assertFalse(match(g, Arguments(int)))
        self.assertTrue(match(h, Arguments(None, int, Remaining(str))))

    def test_keyword_arguments(self):
        # noinspection PyUnusedLocal
        def f(x: int, y: bool):
            pass

        # noinspection PyUnusedLocal
        def g(*, x: int):
            pass

        self.assertTrue(match(f, Arguments(x=int)))
        self.assertFalse(match(f, Strict(Arguments(x=int))))
        self.assertTrue(match(f, Strict(Arguments(x=int, y=bool))))
        self.assertTrue(match(g, Arguments(x=int)))

    def test_returns(self):
        def f() -> int:
            pass

        self.assertTrue(match(f, Returns(int)))
        self.assertFalse(match(f, Returns(float)))

    def test_at(self):
        record = {
            "foo": {
                "bar": {
                    "quux": {
                        "value": "deeply nested"
                    }
                }
            }
        }

        result = match(record, At("foo.bar.quux", {"value": Capture(..., name="value")}))
        self.assertTrue(result)
        self.assertEqual("deeply nested", result['value'])

        result = match(record, At(['foo', 'bar', 'quux'], {"value": Capture(..., name="value")}))
        self.assertTrue(result)
        self.assertEqual("deeply nested", result['value'])

        self.assertFalse(match(record, At("foo.bar.quux.baz", InstanceOf(int))))

    def test_string(self):
        path = "https://somehost/foo=1/bar=2/"
        result = match(path, String(
            Capture(OneOf("https", "http"), name="protocol"),
            "://",
            Capture(Regex("[a-zA-Z_-]+"), name="host"),
            Regex("/+"),
            "foo=",
            Capture(Regex("[^=/]+"), name="foo"),
            "/",
            "bar=",
            Capture(Regex("[^=/]+"), name="bar"),
            Regex("/*"),
        ))
        self.assertTrue(result)
        self.assertEqual("https", result['protocol'])
        self.assertEqual("somehost", result['host'])
        self.assertEqual("1", result['foo'])
        self.assertEqual("2", result['bar'])

    def test_string_argresult(self):
        path = "https://somehost/foo=1/bar=2/"

        result = match(path, String(
            Capture(OneOf("https", "http"), name="protocol"),
            "://",
            Capture(Regex("[a-zA-Z_-]+"), name="host"),
            Regex("/+"),
            "foo=",
            Capture(Regex("[^=/]+"), name="foo"),
            "/",
            "bar=",
            Capture(Regex("[^=/]+"), name="bar"),
            Regex("/*"),
        ))
        self.assertTrue(result)
        self.assertEqual("https", result.protocol)
        self.assertEqual("somehost", result.host)
        self.assertEqual("1", result.foo)
        self.assertEqual("2", result.bar)

    def test_string_with_at_capture(self):
        path = "https://somehost/foo=1/bar=2/"

        result = match(path, String(
            'protocol' @ OneOf("https", "http"),
            "://",
            'host' @ Regex("[a-zA-Z_-]+"),
            Regex("/+"),
            "foo=",
            'foo' @ Regex("[^=/]+"),
            "/",
            "bar=",
            'bar' @ Regex("[^=/]+"),
            Regex("/*"),
        ))
        self.assertTrue(result)
        self.assertEqual("https", result.protocol)
        self.assertEqual("somehost", result.host)
        self.assertEqual("1", result.foo)
        self.assertEqual("2", result.bar)

    def test_is_type(self):
        self.assertTrue(match(1, IsNumber))
        self.assertTrue(match(1.0, IsNumber))
        self.assertFalse(match(True, IsNumber))

    def test_object(self):
        request = {
            "api_version": "v1",
            "job": {
                "run_at": "2020-08-27 14:09:30",
                "command": "echo 'booya'",
            }
        }

        # noinspection PyUnresolvedReferences
        result = match(request, Items(
            api_version="v1",
            job=Items(
                run_at=Transformed(datetime.fromisoformat, 'time' @ _),
            ) & OneOf(
                Items(command='command' @ InstanceOf(str)),
                Items(spawn='container' @ InstanceOf(str)),
            )
        ))
        self.assertTrue(result)
        self.assertFalse('container' in result)
        self.assertEqual(result['time'], datetime(2020, 8, 27, 14, 9, 30))
        self.assertEqual(result['command'], "echo 'booya'")

    def test_attrs(self):
        @dataclass
        class Thing:
            foo: str
            bar: int

        self.assertTrue(match(Thing('xyz', 4711), Attrs(foo='xyz')))
        self.assertFalse(match(Thing('xyz', 4711), Strict(Attrs(foo='xyz'))))
        self.assertTrue(match(Thing('xyz', 4711), Attrs(foo=InstanceOf(str), bar=InstanceOf(int))))
        self.assertTrue(match(Thing('xyz', 4711), Strict(Attrs(foo=InstanceOf(str), bar=InstanceOf(int)))))

    def test_transformed_exception(self):
        request = {
            "api_version": "v1",
            "job": {
                "run_at": "junk",
                "command": "echo 'booya'",
            }
        }

        # noinspection PyUnresolvedReferences
        self.assertFalse(match(request, Items(
            job=Items(
                run_at=Transformed(datetime.fromisoformat, 'time' @ _),
            )
        )))

    def test_splat_result(self):
        def f(**kwargs):
            dct = {}
            for k, v in kwargs.items():
                dct[k] = v
            return dct

        result = match({'a': 1, 'b': 2, 'c': 3}, Items(a=_ >> 'x', b=_ >> 'y', c=_ >> 'z'))
        self.assertTrue(result)
        r = f(**result)
        self.assertEqual({'x': 1, 'y': 2, 'z': 3}, r)

    def test_maybe(self):
        pattern = Items(a=...) & Maybe(Items(b=_ >> 'capture'))

        result = match({'a': 3}, pattern)
        self.assertTrue(result)
        self.assertFalse('capture' in result)
        result = match({'a': 3, 'b': 5}, pattern)
        self.assertTrue(result)
        self.assertTrue('capture' in result)
        self.assertEqual(5, result['capture'])

    def test_dataclasses(self):
        @dataclass
        class User:
            first_name: str
            last_name: str

        @dataclass
        class LoggedInUser(User):
            display_name: str

        result = match(User("Jane", "Doe"), User("Jane", _ >> 'last_name'))
        self.assertTrue(result)
        self.assertEqual("Doe", result['last_name'])

        result = match(LoggedInUser("John", "Doe", "johndoe"), User(_, _ >> 'last_name'))
        self.assertTrue(result)
        self.assertEqual("Doe", result['last_name'])

    def test_bind(self):
        result = {}
        if match({'a': 1, 'b': 2}, {'a': _ >> 'foo'}).bind(result):
            self.assertEqual(1, result['foo'])
        self.assertEqual({'foo': 1}, result)

    def test_extract(self):
        # noinspection PyUnusedLocal
        def f(arg: str) -> float:
            pass

        # noinspection PyUnusedLocal
        def g(a: int, b: int):
            pass

        # noinspection PyUnresolvedReferences
        result = match(f, Arguments('arg_types' @ Many(...)) & Returns('return_type' @ _))
        self.assertTrue(result)
        self.assertEqual([str], result.arg_types)
        self.assertEqual(float, result.return_type)

        self.assertTrue(match(f, Arguments(_)))
        self.assertFalse(match(g, Arguments(_)))
        self.assertFalse(match(f, Arguments(_, _)))
        self.assertTrue(match(g, Arguments(_, _)))

    def test_strict_match(self):
        self.assertTrue(match({'a': 3}, {'a': 3.0}))
        self.assertFalse(match({'a': 3}, {'a': 3.0}, strict=True))

    def test_groups(self):
        # noinspection PyUnresolvedReferences
        result = match([1, 2, 3], ['a' @ _, 'b' @ _, 'c' @ _])
        self.assertTrue(result)
        self.assertEqual(3, len(result))
        self.assertEqual({'a': 1, 'b': 2, 'c': 3}, result.groups())

    def test_subclass_of(self):
        result = match(bool, SubclassOf(int))
        self.assertTrue(result)

    def test_is_truish(self):
        self.assertTrue(match([0], IsTruish))
        self.assertTrue(match([], Not(IsTruish)))
        self.assertFalse(match([], IsTruish))

    def test_range(self):
        self.assertTrue(match([0, 1, 2], range(0, 3)))
        self.assertFalse(match([0, 1, 2], range(0, 3), strict=True))
        self.assertTrue(match(range(0, 3), range(0, 3), strict=True))
        self.assertFalse(match(range(0, 3), range(0, 2), strict=True))
        self.assertFalse(match(range(0, 2), range(0, 3), strict=True))



from .  import good, fail

# dictionary of string -> tuple, with tuple of two elements with different type
# In this case, each value should have the same two types
good('dict(str:tuple(type(x),type(y))),x!=y', {'a': (2, 1.1)})
fail('dict(str:tuple(type(x),type(y))),x!=y', {'a': (2, 1)})

# This fails because we have x=int,y=float followed by float,int
fail('dict(str:tuple(type(x),type(y))),x!=y', {'a': (2, 1.1), 'b': (1.1, 2)})

# Here we force the context to not match using $(...) 
good('dict(str:$(tuple(type(x),type(y)),x!=y))', {'a': (2, 1.1),
                                                  'b': (1.1, 2)})
fail('dict(str:$(tuple(type(x),type(y)),x!=y))', {'a': (2, 1)})



import contextlib
import _imp
import importlib
import importlib.util
import os
import sys
import unittest
import warnings

from .os_helper import unlink


@contextlib.contextmanager
def _ignore_deprecated_imports(ignore=True):
    """Context manager to suppress package and module deprecation
    warnings when importing them.

    If ignore is False, this context manager has no effect.
    """
    if ignore:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", ".+ (module|package)",
                                    DeprecationWarning)
            yield
    else:
        yield


def unload(name):
    try:
        del sys.modules[name]
    except KeyError:
        pass


def forget(modname):
    """'Forget' a module was ever imported.

    This removes the module from sys.modules and deletes any PEP 3147/488 or
    legacy .pyc files.
    """
    unload(modname)
    for dirname in sys.path:
        source = os.path.join(dirname, modname + '.py')
        # It doesn't matter if they exist or not, unlink all possible
        # combinations of PEP 3147/488 and legacy pyc files.
        unlink(source + 'c')
        for opt in ('', 1, 2):
            unlink(importlib.util.cache_from_source(source, optimization=opt))


def make_legacy_pyc(source):
    """Move a PEP 3147/488 pyc file to its legacy pyc location.

    :param source: The file system path to the source file.  The source file
        does not need to exist, however the PEP 3147/488 pyc file must exist.
    :return: The file system path to the legacy pyc file.
    """
    pyc_file = importlib.util.cache_from_source(source)
    up_one = os.path.dirname(os.path.abspath(source))
    legacy_pyc = os.path.join(up_one, source + 'c')
    os.rename(pyc_file, legacy_pyc)
    return legacy_pyc


def import_module(name, deprecated=False, *, required_on=()):
    """Import and return the module to be tested, raising SkipTest if
    it is not available.

    If deprecated is True, any module or package deprecation messages
    will be suppressed. If a module is required on a platform but optional for
    others, set required_on to an iterable of platform prefixes which will be
    compared against sys.platform.
    """
    with _ignore_deprecated_imports(deprecated):
        try:
            return importlib.import_module(name)
        except ImportError as msg:
            if sys.platform.startswith(tuple(required_on)):
                raise
            raise unittest.SkipTest(str(msg))


def _save_and_remove_modules(names):
    orig_modules = {}
    prefixes = tuple(name + '.' for name in names)
    for modname in list(sys.modules):
        if modname in names or modname.startswith(prefixes):
            orig_modules[modname] = sys.modules.pop(modname)
    return orig_modules


# @contextlib.contextmanager
# def frozen_modules(enabled=True):
#     """Force frozen modules to be used (or not).
#
#     This only applies to modules that haven't been imported yet.
#     Also, some essential modules will always be imported frozen.
#     """
#     _imp._override_frozen_modules_for_tests(1 if enabled else -1)
#     try:
#         yield
#     finally:
#         _imp._override_frozen_modules_for_tests(0)


# def import_fresh_module(name, fresh=(), blocked=(), *,
#                         deprecated=False,
#                         usefrozen=False,
#                         ):
#     """Import and return a module, deliberately bypassing sys.modules.
#
#     This function imports and returns a fresh copy of the named Python module
#     by removing the named module from sys.modules before doing the import.
#     Note that unlike reload, the original module is not affected by
#     this operation.
#
#     *fresh* is an iterable of additional module names that are also removed
#     from the sys.modules cache before doing the import. If one of these
#     modules can't be imported, None is returned.
#
#     *blocked* is an iterable of module names that are replaced with None
#     in the module cache during the import to ensure that attempts to import
#     them raise ImportError.
#
#     The named module and any modules named in the *fresh* and *blocked*
#     parameters are saved before starting the import and then reinserted into
#     sys.modules when the fresh import is complete.
#
#     Module and package deprecation messages are suppressed during this import
#     if *deprecated* is True.
#
#     This function will raise ImportError if the named module cannot be
#     imported.
#
#     If "usefrozen" is False (the default) then the frozen importer is
#     disabled (except for essential modules like importlib._bootstrap).
#     """
#     # NOTE: test_heapq, test_json and test_warnings include extra sanity checks
#     # to make sure that this utility function is working as expected
#     with _ignore_deprecated_imports(deprecated):
#         # Keep track of modules saved for later restoration as well
#         # as those which just need a blocking entry removed
#         fresh = list(fresh)
#         blocked = list(blocked)
#         names = {name, *fresh, *blocked}
#         orig_modules = _save_and_remove_modules(names)
#         for modname in blocked:
#             sys.modules[modname] = None
#
#         try:
#             with frozen_modules(usefrozen):
#                 # Return None when one of the "fresh" modules can not be imported.
#                 try:
#                     for modname in fresh:
#                         __import__(modname)
#                 except ImportError:
#                     return None
#                 return importlib.import_module(name)
#         finally:
#             _save_and_remove_modules(names)
#             sys.modules.update(orig_modules)


# class CleanImport(object):
#     """Context manager to force import to return a new module reference.
#
#     This is useful for testing module-level behaviours, such as
#     the emission of a DeprecationWarning on import.
#
#     Use like this:
#
#         with CleanImport("foo"):
#             importlib.import_module("foo") # new reference
#
#     If "usefrozen" is False (the default) then the frozen importer is
#     disabled (except for essential modules like importlib._bootstrap).
#     """
#
#     def __init__(self, *module_names, usefrozen=False):
#         self.original_modules = sys.modules.copy()
#         for module_name in module_names:
#             if module_name in sys.modules:
#                 module = sys.modules[module_name]
#                 # It is possible that module_name is just an alias for
#                 # another module (e.g. stub for modules renamed in 3.x).
#                 # In that case, we also need delete the real module to clear
#                 # the import cache.
#                 if module.__name__ != module_name:
#                     del sys.modules[module.__name__]
#                 del sys.modules[module_name]
#         self._frozen_modules = frozen_modules(usefrozen)
#
#     def __enter__(self):
#         self._frozen_modules.__enter__()
#         return self
#
#     def __exit__(self, *ignore_exc):
#         sys.modules.update(self.original_modules)
#         self._frozen_modules.__exit__(*ignore_exc)


class DirsOnSysPath(object):
    """Context manager to temporarily add directories to sys.path.

    This makes a copy of sys.path, appends any directories given
    as positional arguments, then reverts sys.path to the copied
    settings when the context ends.

    Note that *all* sys.path modifications in the body of the
    context manager, including replacement of the object,
    will be reverted at the end of the block.
    """

    def __init__(self, *paths):
        self.original_value = sys.path[:]
        self.original_object = sys.path
        sys.path.extend(paths)

    def __enter__(self):
        return self

    def __exit__(self, *ignore_exc):
        sys.path = self.original_object
        sys.path[:] = self.original_value


def modules_setup():
    return sys.modules.copy(),


def modules_cleanup(oldmodules):
    # Encoders/decoders are registered permanently within the internal
    # codec cache. If we destroy the corresponding modules their
    # globals will be set to None which will trip up the cached functions.
    encodings = [(k, v) for k, v in sys.modules.items()
                 if k.startswith('encodings.')]
    sys.modules.clear()
    sys.modules.update(encodings)
    # XXX: This kind of problem can affect more than just encodings.
    # In particular extension modules (such as _ssl) don't cope
    # with reloading properly. Really, test modules should be cleaning
    # out the test specific modules they know they added (ala test_runpy)
    # rather than relying on this function (as test_importhooks and test_pkg
    # do currently). Implicitly imported *real* modules should be left alone
    # (see issue 10556).
    sys.modules.update(oldmodules)



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, binding=True
# THIS FILE IS GENERATED - DO NOT EDIT #

from typing import Any
from collections.abc import Generator
import cython  # type: ignore
from cython import (declare as decl, address as addr, sizeof, typeof, struct, 
                    union, typedef, cast, char, short, int as cint, double, long, longlong)

from pyrsistent import (
    pset, 
    pmap, 
    pvector, 
    s, v, m
)
from Aspidites.woma import *
from Aspidites._vendor import (
    take,
    drop,
    takelast,
    droplast,
    consume,
    nth,         
    first_true,  
    iterate,     
    padnone,     
    ncycles,     
    repeatfunc,  
    grouper,     
    group_by,    
    roundrobin,  
    partition,   
    splitat,     
    splitby,     
    powerset,    
    pairwise,    
    iter_suppress,
    flatten,     
    accumulate,  
    reduce,      
    filterfalse, 
    zip_longest,
    call,   
    apply,  
    flip,   
    curry,  
    zipwith,
    foldl,  
    foldr,  
    unfold,
    match as __match,
    _,
)
from Aspidites.monads import Maybe, Surely
from Aspidites.math import Undefined, SafeDiv, SafeExp, SafeMod, SafeFloorDiv, SafeUnaryAdd, SafeUnarySub, SafeFactorial
from Aspidites._vendor.contracts import contract, new_contract
from Aspidites._vendor.RestrictedPython import safe_builtins
safe_builtins['print'] = print
from Aspidites._vendor.RestrictedPython import compile_restricted as compile
safe_builtins['compile'] = compile
# DECLARATIONS TO ALLOW CONTRACTS TO TYPE CHECK #
procedure: None
coroutine: Generator
number: Any
globals().update(dict(__builtins__=safe_builtins))  # add all imports to globals


@contract()
def add(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return x+y


@contract()
def sub(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return x-y


@contract()
def div(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return Maybe(SafeDiv, x, y)()


@contract()
def exp(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return Maybe(SafeExp, x, y, )()


@contract()
def mod(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return Maybe(SafeMod, x, y, )()





from Aspidites._vendor.contracts import contract, new_contract, ContractNotRespected
import unittest


class ClassContractsTests(unittest.TestCase):

    def test_class_contract1(self):

        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            @new_contract
            def legal_move1(self, move):
                return move in self.legal

            @contract(move='legal_move1')
            def take_turn(self, move):
                pass

        g1 = Game([1, 2])
        g1.take_turn(1)
        g1.take_turn(2)
        self.assertRaises(ContractNotRespected, g1.take_turn, 3)

    def test_class_contract2(self):

        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            @new_contract
            def legal_move2(self, move):
                return move in self.legal

            @contract(move='legal_move2')
            def take_turn(self, move):
                pass

        g1 = Game([1, 2])
        g1.take_turn(1)
        g1.take_turn(2)
        self.assertRaises(ContractNotRespected, g1.take_turn, 3)

    def test_class_contract3(self):

        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            def legal_move(self, move):
                return move in self.legal

            new_contract('alegalmove', legal_move)

            @contract(move='alegalmove')
            def take_turn(self, move):
                pass

        g1 = Game([1, 2])
        g1.take_turn(1)
        g1.take_turn(2)
        self.assertRaises(ContractNotRespected, g1.take_turn, 3)

    def test_class_contract1_bad(self):
        """ example of bad usage, using the contract from outside """

        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            @new_contract
            def legal_move4(self, move):
                return move in self.legal

        def go():
            @contract(move='legal_move4')
            def take_turn(move):
                pass

            take_turn(0)

        self.assertRaises(ContractNotRespected, go)






#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from Aspidites._vendor.contracts.main import new_contract
from math import isfinite

try:
    import numpy  # @UnusedImport
except ImportError:  # pragma: no cover
    __all__ = ['finite']
    new_contract('float', 'Float')
    new_contract('int', 'Int')
    new_contract('number', 'float|int')

    @new_contract
    def finite(x):
        return isfinite(x)
else:
    __all__ = []
    new_contract('float', 'Float|np_scalar_float|(np_scalar, array(float))')
    new_contract('int', 'Int|np_scalar_int|(np_scalar,array(int))')
    new_contract('uint', 'np_scalar_uint|(np_scalar, array(uint))')
    new_contract('number', 'float|int|uint')






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) The python-semanticversion project
# This code is distributed under the two-clause BSD License.

import sys
import unittest

import pytest

pytestmark = pytest.mark.filterwarnings(
    "ignore::DeprecationWarning", "ignore::PendingDeprecationWarning"
)
from Aspidites._vendor.semantic_version import Spec, SpecItem, Version, match


class MatchTestCase(unittest.TestCase):
    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

    invalid_specs = [
        "",
        "!0.1",
        "<=0.1.4a",
        ">0.1.1.1",
        "<0.1.2-rc1.3-14.15+build.2012-01-01.11h34",
    ]

    valid_specs = [
        "*",
        "==0.1.0",
        "=0.1.0",
        "0.1.0",
        "<=0.1.1",
        "<0.1",
        "1",
        ">0.1.2-rc1",
        ">=0.1.2-rc1.3.4",
        "==0.1.2+build42-12.2012-01-01.12h23",
        "!=0.1.2-rc1.3-14.15+build.2012-01-01.11h34",
        "^0.1.2",
        "~0.1.2",
        "~=0.1.2",
    ]

    matches = {
        "*": [
            "0.1.1",
            "0.1.1+build4.5",
            "0.1.2-rc1",
            "0.1.2-rc1.3",
            "0.1.2-rc1.3.4",
            "0.1.2+build42-12.2012-01-01.12h23",
            "0.1.2-rc1.3-14.15+build.2012-01-01.11h34",
            "0.2.0",
            "1.0.0",
        ],
        "==0.1.2": [
            "0.1.2+build42-12.2012-01-01.12h23",
        ],
        "=0.1.2": [
            "0.1.2+build42-12.2012-01-01.12h23",
        ],
        "0.1.2": [
            "0.1.2+build42-12.2012-01-01.12h23",
        ],
        "<=0.1.2": [
            "0.1.1",
            "0.1.2-rc1",
            "0.1.2-rc1.3.4",
            "0.1.2",
            "0.1.2+build4",
        ],
        "!=0.1.2+": [
            "0.1.2+1",
            "0.1.2-rc1",
        ],
        "!=0.1.2-": [
            "0.1.1",
            "0.1.2-rc1",
        ],
        "!=0.1.2+345": [
            "0.1.1",
            "0.1.2-rc1+345",
            "0.1.2+346",
            "0.2.3+345",
        ],
        ">=0.1.1": [
            "0.1.1",
            "0.1.1+build4.5",
            "0.1.2-rc1.3",
            "0.2.0",
            "1.0.0",
        ],
        ">0.1.1": [
            "0.1.2+build4.5",
            "0.1.2-rc1.3",
            "0.2.0",
            "1.0.0",
        ],
        "<0.1.1-": [
            "0.1.1-alpha",
            "0.1.1-rc4",
            "0.1.0+12.3",
        ],
        "^0.1.2": [
            "0.1.2",
            "0.1.2+build4.5",
            "0.1.3-rc1.3",
            "0.1.4",
        ],
        "~0.1.2": [
            "0.1.2",
            "0.1.2+build4.5",
            "0.1.3-rc1.3",
        ],
        "~=1.4.5": (
            "1.4.5",
            "1.4.10-alpha",
            "1.4.10",
        ),
        "~=1.4": [
            "1.4.0",
            "1.6.10-alpha",
            "1.6.10",
        ],
    }

    def test_invalid(self):
        for invalid in self.invalid_specs:
            with self.subTest(spec=invalid):
                with self.assertRaises(
                    ValueError, msg="Spec(%r) should be invalid" % invalid
                ):
                    Spec(invalid)

    def test_simple(self):
        for valid in self.valid_specs:
            with self.subTest(spec=valid):
                spec = SpecItem(valid)
                normalized = str(spec)
                self.assertEqual(spec, SpecItem(normalized))

    def test_match(self):
        for spec_text, versions in self.matches.items():
            for version_text in versions:
                with self.subTest(spec=spec_text, version=version_text):
                    spec = Spec(spec_text)
                    self.assertNotEqual(spec, spec_text)
                    version = Version(version_text)
                    self.assertIn(version, spec)
                    self.assertTrue(
                        spec.match(version), "%r does not match %r" % (version, spec)
                    )
                    self.assertTrue(match(spec_text, version_text))

    def test_contains(self):
        spec = Spec("<=0.1.1")
        self.assertFalse("0.1.0" in spec, "0.1.0 should not be in %r" % spec)

        version = Version("0.1.1+4.2")
        self.assertTrue(version in spec, "%r should be in %r" % (version, spec))

        version = Version("0.1.1-rc1+4.2")
        self.assertTrue(version in spec, "%r should be in %r" % (version, spec))

    def test_prerelease_check(self):
        strict_spec = Spec(">=0.1.1-")
        lax_spec = Spec(">=0.1.1")
        version = Version("0.1.1-rc1+4.2")
        self.assertFalse(
            version in lax_spec, "%r should not be in %r" % (version, lax_spec)
        )
        self.assertFalse(
            version in strict_spec, "%r should not be in %r" % (version, strict_spec)
        )

    def test_build_check(self):
        spec = Spec("<=0.1.1-rc1")
        version = Version("0.1.1-rc1+4.2")
        self.assertTrue(version in spec, "%r should be in %r" % (version, spec))


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class CaseDistinctionTest(unittest.TestCase):

    def test_case_distinction(self):
        @case_distinction
        def fib(n: Match(OneOf(0, 1))):
            return n

        @case_distinction
        def fib(n):
            return fib(n - 2) + fib(n - 1)

        self.assertEqual(0, fib(0))
        self.assertEqual(1, fib(1))
        self.assertEqual(1, fib(2))
        self.assertEqual(2, fib(3))
        self.assertEqual(3, fib(4))
        self.assertEqual(5, fib(5))
        self.assertEqual(8, fib(6))

    def test_readme_example(self):
        value = 7

        def check(v):
            self.assertEqual("It's between 1 and 10", v)

        @case_distinction
        def f(n: Match(Between(1, 10))):
            check("It's between 1 and 10")
            return 1

        @case_distinction
        def f(n: Match(Between(11, 20))):
            check("It's between 11 and 20")
            return 2

        @case_distinction
        def f(n):
            check("It's not between 1 and 20")
            return 3

        self.assertEqual(1, f(value))

    def test_overloading(self):
        def check(v):
            pass

        @case_distinction
        def f(a: str, b: str):
            return 0

        @case_distinction
        def f(a: int, b: int):
            return 1

        @case_distinction
        def f(a: int, b: int, c: int):
            return 2

        self.assertEqual(0, f("a", "b"))
        self.assertEqual(1, f(1, 2))
        self.assertEqual(2, f(1, 2, 3))



from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class Glob(unittest.TestCase):

    def test_empty(self):
        self.assertTrue(match(
            [1, 2, 3],
            [1, Many(...), 2, 3]
        ))

    def test_at_least(self):
        self.assertTrue(match(
            [1, 2, 3],
            [
                1,
                Many(..., at_least=2),
            ]))
        self.assertFalse(match(
            [1, 2],
            [
                1,
                Many(..., at_least=2),
            ]))

    def test_at_most(self):
        self.assertTrue(match(
            [1, 2, 3, 4],
            [
                1,
                Many(..., at_most=2),
                4,
            ]))
        self.assertFalse(match(
            [1, 2, 3],
            [
                1,
                Many(..., at_most=2),
                4,
            ]))

    def test_with_remaining(self):
        self.assertTrue(match(
            [1, 2, 3],
            [
                1,
                Many(..., at_most=2),
                Remaining(...),
            ]))

    def test_multiple(self):
        self.assertTrue(match(
            range(1, 10),
            [
                1,
                2,
                Many(...),
                6,
                Many(...),
                8,
                9,
            ]
        ))

    def test_capture(self):
        result = match(
            range(1, 10),
            [
                1,
                2,
                Capture(Many(...), name=1),
                6,
                Capture(Many(...), name=2),
                8,
                9,
            ]
        )
        self.assertTrue(result)
        self.assertEqual(result[1], [3, 4, 5])
        self.assertEqual(result[2], [7])

    def test_exactly(self):
        result = match(range(1, 10), [
            1,
            Many(..., exactly=3) >> 'a',
            Many(...) >> 'b',
            6,
            7,
            8,
            9
        ])
        self.assertTrue(result)
        self.assertEqual([2, 3, 4], result.a)
        self.assertEqual([5], result.b)

    def test_exactly_empty_following(self):
        result = match(range(1, 10), [
            1,
            Many(..., exactly=4) >> 'a',
            Many(...) >> 'b',
            6,
            7,
            8,
            9
        ])
        self.assertTrue(result)
        self.assertEqual([2, 3, 4, 5], result.a)
        self.assertEqual([], result.b)
        self.assertEqual("MatchResult(matches=True, groups={'a': [2, 3, 4, 5], 'b': []})", repr(result))



# shapes.py
#
#   A sample program showing how parse actions can convert parsed
# strings into a data type or object.
#
# Copyright 2012, Paul T. McGuire
#

# define class hierarchy of Shape classes, with polymorphic area method
class Shape(object):
    def __init__(self, tokens):
        self.__dict__.update(tokens.asDict())

    def area(self):
        raise NotImplementedException()

    def __str__(self):
        return "<{0}>: {1}".format(self.__class__.__name__, self.__dict__)

class Square(Shape):
    def area(self):
        return self.side**2

class Rectangle(Shape):
    def area(self):
        return self.width * self.height

class Circle(Shape):
    def area(self):
        return 3.14159 * self.radius**2


from pyparsing import *

number = Regex(r'-?\d+(\.\d*)?').setParseAction(lambda t:float(t[0]))

# Shape expressions:
#   square : S <centerx> <centery> <side>
#   rectangle: R <centerx> <centery> <width> <height>
#   circle : C <centerx> <centery> <diameter>

squareDefn = "S" + number('centerx') + number('centery') + number('side')
rectDefn = "R" + number('centerx') + number('centery') + number('width') + number('height')
circleDefn = "C" + number('centerx') + number('centery') + number('diameter')

squareDefn.setParseAction(Square)
rectDefn.setParseAction(Rectangle)

def computeRadius(tokens):
    tokens['radius'] = tokens.diameter/2.0
circleDefn.setParseAction(computeRadius, Circle)

shapeExpr = squareDefn | rectDefn | circleDefn

tests = """\
C 0 0 100
R 10 10 20 50
S -1 5 10""".splitlines()

for t in tests:
    shape = shapeExpr.parseString(t)[0]
    print(shape)
    print("Area:", shape.area())
    print()



# parseListString.py
#
# Copyright, 2006, by Paul McGuire
#

from pyparsing import *

# first pass
lbrack = Literal("[")
rbrack = Literal("]")
integer = Word(nums).setName("integer")
real = Combine(Optional(oneOf("+ -")) + Word(nums) + "." +
               Optional(Word(nums))).setName("real")

listItem = real | integer | quotedString

listStr = lbrack + delimitedList(listItem) + rbrack

test = "['a', 100, 3.14]"

print(listStr.parseString(test))


# second pass, cleanup and add converters
lbrack = Literal("[").suppress()
rbrack = Literal("]").suppress()
cvtInt = lambda s,l,toks: int(toks[0])
integer = Word(nums).setName("integer").setParseAction( cvtInt )
cvtReal = lambda s,l,toks: float(toks[0])
real = Regex(r'[+-]?\d+\.\d*').setName("floating-point number").setParseAction( cvtReal )
listItem = real | integer | quotedString.setParseAction( removeQuotes )

listStr = lbrack + delimitedList(listItem) + rbrack

test = "['a', 100, 3.14]"

print(listStr.parseString(test))

# third pass, add nested list support, and tuples, too!
cvtInt = lambda s,l,toks: int(toks[0])
cvtReal = lambda s,l,toks: float(toks[0])

lbrack = Literal("[").suppress()
rbrack = Literal("]").suppress()
integer = Word(nums).setName("integer").setParseAction( cvtInt )
real = Regex(r'[+-]?\d+\.\d*').setName("floating-point number").setParseAction( cvtReal )
tupleStr = Forward()
listStr = Forward()
listItem = real | integer | quotedString.setParseAction(removeQuotes) | Group(listStr) | tupleStr
tupleStr << ( Suppress("(") + delimitedList(listItem) + Optional(Suppress(",")) + Suppress(")") )
tupleStr.setParseAction( lambda t:tuple(t.asList()) )
listStr << lbrack + delimitedList(listItem) + Optional(Suppress(",")) + rbrack

test = "['a', 100, ('A', [101,102]), 3.14, [ +2.718, 'xyzzy', -1.414] ]"
print(listStr.parseString(test))

# fourth pass, add parsing of dicts
cvtInt = lambda s,l,toks: int(toks[0])
cvtReal = lambda s,l,toks: float(toks[0])
cvtDict = lambda s,l,toks: dict(toks[0])

lbrack = Literal("[").suppress()
rbrack = Literal("]").suppress()
lbrace = Literal("{").suppress()
rbrace = Literal("}").suppress()
colon = Literal(":").suppress()
integer = Word(nums).setName("integer").setParseAction( cvtInt )
real = Regex(r'[+-]?\d+\.\d*').setName("real").setParseAction( cvtReal )

tupleStr = Forward()
listStr = Forward()
dictStr = Forward()
listItem = real | integer | quotedString.setParseAction(removeQuotes) | Group(listStr) | tupleStr | dictStr
tupleStr <<= ( Suppress("(") + delimitedList(listItem) + Optional(Suppress(",")) + Suppress(")") )
tupleStr.setParseAction( lambda t:tuple(t.asList()) )
listStr <<= (lbrack + Optional(delimitedList(listItem)) + Optional(Suppress(",")) + rbrack)
dictKeyStr = real | integer | quotedString.setParseAction(removeQuotes)
dictStr <<= lbrace + Optional(delimitedList( Group( dictKeyStr + colon + listItem ))) + Optional(Suppress(",")) + rbrace
dictStr.setParseAction(lambda t: {k_v[0]:(k_v[1].asList() if isinstance(k_v[1],ParseResults) else k_v[1]) for k_v in t})

test = '[{0: [2], 1: []}, {0: [], 1: [], 2: [,]}, {0: [1, 2,],}]'
print(listStr.parseString(test))



import unittest

from Aspidites._vendor.apm import *


# noinspection PyUnusedLocal
def f(x: int, *xs: int, z: str = None, **kwargs):
    pass


# noinspection PyUnusedLocal
def g(a: str, m: float, n: float, z: int):
    pass


class TypeFooTest(unittest.TestCase):
    def test_parameters(self):
        self.assertTrue(match(f, Parameters(_, _, _, z=_)))
        self.assertFalse(match(f, Parameters(_, _, _, x=_)))

    def test_parameter_patterns(self):
        self.assertTrue(match(f, Parameters(int, VarArgs, KwArgs)))
        self.assertFalse(match(f, Parameters(int, KwArgs)))
        self.assertTrue(match(f, Parameters(int, VarArgs(int), KwArgs)))
        self.assertFalse(match(f, Parameters(int, VarArgs(float), KwArgs)))
        self.assertFalse(match(g, Parameters(str, float, float, VarArgs(int))))

    def test_some(self):
        pat = Parameters(str, Many(float), int)
        self.assertTrue(match(g, pat))
        self.assertFalse(match(f, pat))


if __name__ == '__main__':
    unittest.main()





__all__ = ["F",  # BEGIN func.py
            "curried",  # END func.py
            "take",  # BEGIN iters.py
            "drop",
            "takelast",
            "droplast",
            "consume",
            "nth",
            "first_true",
            "iterate",
            "padnone",
            "ncycles",
            "repeatfunc",
            "grouper",
            "group_by",
            "roundrobin",
            "partition",
            "splitat",
            "splitby",
            "powerset",
            "pairwise",
            "iter_suppress",
            "flatten",
            "accumulate",  # END iters.py
            "call",  # BEGIN op.py
            "apply",
            "flip",
            "curry",
            "zipwith",
            "foldl",
            "foldr",
            "unfold",  # END op.py
            "tco",
            "stackless",
            "Stream",
            "_",
            "reduce",
            "filterfalse",
            "zip_longest",
            "match",
            "InstanceOf"
            ]
from .fn import *
from .apm import match, InstanceOf


from Aspidites._vendor.RestrictedPython import compile_restricted_exec


def test_call_breakpoint():
    """The Python3.7+ builtin function breakpoint should not
    be used and is forbidden in Aspidites.features.RestrictedPython.
    """
    result = compile_restricted_exec('breakpoint()')
    assert result.errors == ('Line 1: "breakpoint" is a reserved name.',)
    assert result.code is None



import inspect
import typing
from typing import Callable, Dict, List

from .error import MatchError
from .match import match


class Match:
    def __init__(self, pattern, **kwargs):
        self._pattern = pattern
        self._kwargs = kwargs

    def __call__(self, value):
        # typing.get_type_hints requires annotations to be callable, otherwise it bombs out with a TypeError
        return match(value, self._pattern, **self._kwargs)


# noinspection PyDefaultArgument
def overload(fn: Callable, func_map: Dict[str, List[Callable]] = {}):
    qualified_name = fn.__qualname__
    if qualified_name not in func_map:
        func_map[qualified_name] = []
    cases = func_map[qualified_name]
    cases.append(fn)

    def wrapper(*args, **kwargs):
        for func in func_map[qualified_name]:
            sig: inspect.Signature = inspect.signature(func)
            try:
                bound: inspect.BoundArguments = sig.bind(*args, **kwargs)
            except TypeError:
                continue
            matches = True
            # resolve the annotations which are strings as per __future__
            type_hints = typing.get_type_hints(func)
            for name, value in bound.arguments.items():
                if name not in type_hints:
                    continue
                annotation = type_hints[name]
                if isinstance(annotation, Match):
                    if not annotation(value):
                        matches = False
                        break
                try:
                    # noinspection PyTypeHints
                    if not isinstance(value, annotation):
                        matches = False
                        break
                except TypeError:
                    pass
            if not matches:
                continue
            return func(*bound.args, **bound.kwargs)
        raise MatchError(f"No match(args={repr(args)}, kwargs={repr(kwargs)})")

    return wrapper


def case_distinction(f):
    return overload(f)



#
# include_preprocessor.py
#
# Short pyparsing script to perform #include inclusions similar to the C preprocessor
#
# Copyright 2019, Paul McGuire
#
from ....._vendor import pyparsing as pp
from pathlib import Path

# parser elements to be used to assemble into #include parser
SEMI = pp.Suppress(';')
INCLUDE = pp.Keyword("#include")
quoted_string = pp.quotedString.addParseAction(pp.removeQuotes)
file_ref = (quoted_string
            | pp.Word(pp.printables, excludeChars=';'))

# parser for parsing "#include xyz.dat;" directives
include_directive = (INCLUDE + file_ref("include_file_name") + SEMI)

# add parse action that will recursively pull in included files - when
# using transformString, the value returned from the parse action will replace
# the text matched by the attached expression
seen = set()
def read_include_contents(s, l, t):
    include_file_ref = t.include_file_name
    include_echo = "/* {} */".format(pp.line(l, s).strip())

    # guard against recursive includes
    if include_file_ref not in seen:
        seen.add(include_file_ref)
        included_file_contents = Path(include_file_ref).read_text()
        return (include_echo + '\n'
                + include_directive.transformString(included_file_contents))
    else:
        lead = ' '*(pp.col(l, s) - 1)
        return "/* recursive include! */\n{}{}".format(lead, include_echo)

# attach include processing method as parse action (parse-time callback)
# to include_directive expression
include_directive.addParseAction(read_include_contents)


if __name__ == '__main__':

    # demo

    # create test files:
    # - a.txt includes b.txt
    # - b.txt includes c.txt
    # - c.txt includes b.txt (must catch infinite recursion)
    Path('a.txt').write_text("""\
        /* a.txt */
        int i;
    
        /* sometimes included files aren't in quotes */    
        #include b.txt;
        """)

    Path('b.txt').write_text("""\
        i = 100;
        #include 'c.txt';
        """)

    Path('c.txt').write_text("""\
        i += 1;
        
        /* watch out! this might be recursive if this file included by b.txt */
        #include b.txt;
        """)


    # use include_directive.transformString to perform includes

    # read contents of original file
    initial_file = Path('a.txt').read_text()

    # print original file
    print(initial_file)
    print('-----------------')

    # expand includes in source file (and any included files) and print the result
    expanded_source = include_directive.transformString(initial_file)
    print(expanded_source)

    # clean up
    for fname in "a.txt b.txt c.txt".split():
        Path(fname).unlink()



#
# gen_ctypes.py
#
# Parse a .h header file to generate ctypes argtype and return type definitions
#
# Copyright 2004-2016, by Paul McGuire
#
from ....._vendor.pyparsing import *

typemap = {
    "byte" : "c_byte",
    "char" : "c_char",
    "char *" : "c_char_p",
    "double" : "c_double",
    "float" : "c_float",
    "int" : "c_int",
    "int16" : "c_int16",
    "int32" : "c_int32",
    "int64" : "c_int64",
    "int8" : "c_int8",
    "long" : "c_long",
    "longlong" : "c_longlong",
    "short" : "c_short",
    "size_t" : "c_size_t",
    "ubyte" : "c_ubyte",
    "uchar" : "c_ubyte",
    "u_char" : "c_ubyte",
    "uint" : "c_uint",
    "u_int" : "c_uint",
    "uint16" : "c_uint16",
    "uint32" : "c_uint32",
    "uint64" : "c_uint64",
    "uint8" : "c_uint8",
    "u_long" : "c_ulong",
    "ulong" : "c_ulong",
    "ulonglong" : "c_ulonglong",
    "ushort" : "c_ushort",
    "u_short" : "c_ushort",
    "void *" : "c_void_p",
    "voidp" : "c_voidp",
    "wchar" : "c_wchar",
    "wchar *" : "c_wchar_p",
    "Bool" : "c_bool",
    "void" : "None",
    }

LPAR,RPAR,LBRACE,RBRACE,COMMA,SEMI = map(Suppress,"(){},;")
ident = Word(alphas, alphanums + "_")
integer = Regex(r"[+-]?\d+")
hexinteger = Regex(r"0x[0-9a-fA-F]+")

const = Suppress("const")
primitiveType = oneOf(t for t in typemap if not t.endswith("*"))
structType = Suppress("struct") + ident
vartype = (Optional(const) +
            (primitiveType | structType | ident) +
            Optional(Word("*")("ptr")))
def normalizetype(t):
    if isinstance(t, ParseResults):
        return ' '.join(t)
        #~ ret = ParseResults([' '.join(t)])
        #~ return ret

vartype.setParseAction(normalizetype)

arg = Group(vartype("argtype") + Optional(ident("argname")))
func_def = (vartype("fn_type") + ident("fn_name") +
                LPAR + Optional(delimitedList(arg|"..."))("fn_args") + RPAR + SEMI)
def derivefields(t):
    if t.fn_args and t.fn_args[-1] == "...":
        t["varargs"]=True
func_def.setParseAction(derivefields)

fn_typedef = "typedef" + func_def
var_typedef = "typedef" + primitiveType("primType") + ident("name") + SEMI

enum_def = (Keyword("enum") + LBRACE +
            delimitedList(Group(ident("name") + '=' + (hexinteger|integer)("value")))("evalues")
            + Optional(COMMA)
            + RBRACE)

c_header = open("snmp_api.h").read()


module = "pynetsnmp"

user_defined_types = set()
typedefs = []
fn_typedefs = []
functions = []
enum_constants = []

# add structures commonly included from std lib headers
def addStdType(t,namespace=""):
    fullname = namespace+'_'+t if namespace else t
    typemap[t] = fullname
    user_defined_types.add(t)
addStdType("fd_set", "sys_select")
addStdType("timeval", "sys_time")

def getUDType(typestr):
    key = typestr.rstrip(" *")
    if key not in typemap:
        user_defined_types.add(key)
        typemap[key] = "{0}_{1}".format(module, key)

def typeAsCtypes(typestr):
    if typestr in typemap:
        return typemap[typestr]
    if typestr.endswith("*"):
        return "POINTER(%s)" % typeAsCtypes(typestr.rstrip(" *"))
    return typestr

# scan input header text for primitive typedefs
for td,_,_ in var_typedef.scanString(c_header):
    typedefs.append( (td.name, td.primType) )
    # add typedef type to typemap to map to itself
    typemap[td.name] = td.name

# scan input header text for function typedefs
fn_typedefs = fn_typedef.searchString(c_header)
# add each function typedef to typemap to map to itself
for fntd in fn_typedefs:
    typemap[fntd.fn_name] = fntd.fn_name

# scan input header text, and keep running list of user-defined types
for fn,_,_ in (cStyleComment.suppress() | fn_typedef.suppress() | func_def).scanString(c_header):
    if not fn: continue
    getUDType(fn.fn_type)
    for arg in fn.fn_args:
        if arg != "...":
            if arg.argtype not in typemap:
                getUDType(arg.argtype)
    functions.append(fn)

# scan input header text for enums
enum_def.ignore(cppStyleComment)
for en_,_,_ in enum_def.scanString(c_header):
    for ev in en_.evalues:
        enum_constants.append( (ev.name, ev.value) )

print("from ctypes import *")
print("{0} = CDLL('{1}.dll')".format(module, module))
print()
print("# user defined types")
for tdname,tdtyp in typedefs:
    print("{0} = {1}".format(tdname, typemap[tdtyp]))
for fntd in fn_typedefs:
    print("{0} = CFUNCTYPE({1})".format(fntd.fn_name,
        ',\n    '.join(typeAsCtypes(a.argtype) for a in fntd.fn_args)))
for udtype in user_defined_types:
    print("class %s(Structure): pass" % typemap[udtype])

print()
print("# constant definitions")
for en,ev in enum_constants:
    print("{0} = {1}".format(en,ev))

print()
print("# functions")
for fn in functions:
    prefix = "{0}.{1}".format(module, fn.fn_name)

    print("{0}.restype = {1}".format(prefix, typeAsCtypes(fn.fn_type)))
    if fn.varargs:
        print("# warning - %s takes variable argument list" % prefix)
        del fn.fn_args[-1]

    if fn.fn_args.asList() != [['void']]:
        print("{0}.argtypes = ({1},)".format(prefix, ','.join(typeAsCtypes(a.argtype) for a in fn.fn_args)))
    else:
        print("%s.argtypes = ()" % (prefix))
